/*
 * Check the positivity of the minimum cycle mean
 * for arbitrary disk weights
 *
 * $Id: check_pos.c,v 1.3 2020/10/07 21:35:27 bousch Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>

/* Portability notes: we assume sizeof(short) >= 2, sizeof(int) >= 4 */

#define	DIRTY1		0x8000
#define	DIRTY2		0x4000
#define	DATAMASK	0x3FFF

#define EMPTY		999

char *Usage = "Usage: %s W[0] W[1] ... W[n-1]\n"
"  where the number of arguments (n) is the number of disks (1<=n<=16)\n"
"  in the cyclic puzzle, and the W[i] are the weights of the various\n"
"  disks (from smallest to greatest), with abs(W[i]) <= 32767.\n"
"The program will say whether the minimum cycle time is >= 0 or < 0.\n";

int Pow4 (int n)
{
	/* This is 4**n */
	return 1 << (n+n);
}

int main (int argc, char **argv)
{
	int D, N, weights[16];
	int pushback, offset, weight, oldval, newval;
	int augmented, dirtybits, minp, maxp;
	int i, n, x, r, r2, vertex, newv, pos[16], smallest[4];
	unsigned short *P;

	D = argc-1;
	if (D<1 || D>16) {
usage_error:
		fprintf(stderr, Usage, argv[0]);
		return 2;
	}
	for (i=0; i<D; i++) {
		weight = atoi(argv[1+i]);
		if (weight < -32767 || weight > 32767)
			goto usage_error;
		weights[i] = weight;
	}
	printf("Disks=%d\n", D);
	printf("Weights:");
	for (i=0; i<D; i++)
		printf(" %d:%d", i, weights[i]);
	printf("\n");
	N = Pow4(D-1);
	printf("Vertices: %d\n", N);
	P = calloc(N, sizeof(unsigned short));
	if (P == NULL) {
		fprintf(stderr, "calloc failed\n");
		return 4;
	}
	/* Mark initially all vertices as dirty */
	for (vertex=0; vertex<N; vertex++)
		P[vertex] = DIRTY1;
	/*
	 * When the largest disk moves, the position must be "pushed back"
	 * by: 1+4+4^2+...+4^{D-1} = (4^D-1)/(4-1).
	 * Beware of overflows when D=16.
	 */
	pushback = 0;
	for (i=0; i<D; i++)
		pushback = 1+4*pushback;
	printf("pushback=%d\n", pushback);
main_loop:
	augmented = 0;
	for (vertex=0; vertex<N; vertex++) {
		if ((P[vertex] & DIRTY1) == 0)
			continue;
		/*
		 * Otherwise, the value of P[vertex] has been modified,
		 * either in the previous loop or the current one.
		 *
		 * Find the smallest disk in each column
		 */
		for (r=0; r<4; r++)
			smallest[r] = EMPTY;	/* A very large number */
		x = vertex;
		for (i=0; i<D; i++) {
			/*
			 * The position of disk i is given by bits 2i and 2i+1
			 * in "vertex". So, we extract the two lowest bits
			 * from x, and shift it.
			 */
			r = x & 03;
			x >>= 2;
			/* Disk i is on peg r */
			pos[i] = r;
			if (i < smallest[r])
				smallest[r] = i;
		}
		/* Find all outgoing edges from this vertex */
		for (r=0; r<4; r++) {
			r2 = (r+1) & 03;
			/* Can we move a disk from peg r to peg r2 ? */
			n = smallest[r];
			if (n < smallest[r2]) {
				/* Disk n on peg r can be moved */
				offset = (r2-r) * Pow4(n);
				newv = vertex + offset;
				weight = weights[n];
				/*
				 * Largest disk moved?
				 * Then translate the new configuration newv
				 * by -1, so that disk D-1 falls back
				 * in column 0
				 */
				if (n == D-1)
					newv -= pushback;
#if 0
				fprintf(stderr, "move %d->%d weight %d\n",
				  vertex, newv, weight);
#endif
				oldval = (P[newv] & DATAMASK);
				newval = (P[vertex] & DATAMASK) - weight;
				/* Will it fit into 14 bits? */
				if (newval > DATAMASK) {
					fprintf(stderr,
					  "Integer overflow, aborting.\n");
					return 8;
				}
				if (newval > oldval) {
					P[newv] = (newval | DIRTY1 | DIRTY2);
					if (augmented < 2000000000)
						++augmented;
				}
			}
		}
	}
	minp = DATAMASK;
	maxp = 0;
	for (vertex=0; vertex<N; vertex++) {
		newval = P[vertex];
		/* Shift the dirty bits, from DIRTY2 to DIRTY1 */
		dirtybits = (newval & DIRTY2);
		newval = (newval & DATAMASK);
		dirtybits = (dirtybits? DIRTY1 : 0);
		P[vertex] = (newval | dirtybits);
		/* And update the minimum and maximum */
		if (newval < minp)
			minp = newval;
		if (newval > maxp)
			maxp = newval;
	}
	printf("min(P)=%d, max(P)=%d, aug=%d\n", minp, maxp, augmented);
	if (minp > 0) {
		printf("No solutions: minimum cycle mean is < 0\n");
		return 1;
	}
	if (augmented)
		goto main_loop;

	/* No augmentation: the array P is a solution */
	printf("Solution found!\n");
	printf("Minimum cycle mean is >= 0\n");

	printf("Theorem: ");
	for (i=0; i<D; i++) {
		weight = weights[i];
		if (weight >= 0)
			putchar('+');
		printf("%d.x%d ", weight, i);
	}
	printf(">= 0.\n");
	return 0;
}
