#!/usr/bin/env python
#
# $Id: hgprof.py,v 1.8 2020/10/31 21:01:01 bousch Exp $
#
# Check that all cycles have positive weight, in the weighted
# Cyclic Tower of Hanoi graph, with weights eventually constant
# and an arbitrary number of disks

import string, sys, os

# List of all height profiles (ie linear orders) on C={0,1,2,3}.
# For each profile, the elements of C are written in ascending
# order, so for instance "2013" means 2 < 0 < 1 < 3.

P4 = [	"0123", "0132", "0213", "0231", "0312", "0321",
	"1023", "1032", "1203", "1230", "1302", "1320",
	"2013", "2031", "2103", "2130", "2301", "2310",
	"3012", "3021", "3102", "3120", "3201", "3210"	]

# Find the possible successors of profile hgp, when a disk is
# moved from column src to column dst.
#
# The initial profile (hgp) is, as a string,
#
#  +--------+---+--------+---+-------+
#  |   s1   |src|  s2a   |dst|  s2b  |
#  +--------+---+--------+---+-------+
#
# and the final profile is
#
#  +--------+---+-------+---+--------+
#  |   s1   |dst|  s2c  |src|  s2d   |
#  +--------+---+-------+---+--------+
#
# with the constraint that s2a+s2b = s2c+s2d.

def transform_hgprof(hgp, src, dst):
  transformed = [ ]
  ps = string.find(hgp, src)
  pd = string.find(hgp, dst)
  if (ps<0) or (pd<0) or (ps>=pd):
    return transformed
  s1 = hgp[:ps]
  s2 = hgp[ps+1:pd] + hgp[pd+1:]
  for i in range(len(s2)+1):
    s3 = s1 + dst + s2[:i] + src + s2[i:]
    transformed.append(s3)
  return transformed

# Special case: the Cyclic Tower of Hanoi on 4 pegs
#
# tr_prof[hgp][c] contains the list of possible successors
# of hgp when a disk is moved from column c to c+1 mod 4

tr_prof = [ ]
for hgp in range(24):
  line = [ ]
  for c in range(4):
    d = (c+1) % 4
    ltr = transform_hgprof(P4[hgp], str(c), str(d))
    next = [ ]
    for tr in ltr:
      hgp2 = P4.index(tr)
      next.append(hgp2)
    line.append(next)
  tr_prof.append(line)

# Command line processing. Special case: launch the program with
# the unique argument "--debug-profiles" to verify that height
# profiles are correctly transformed.

if len(sys.argv) == 2 and sys.argv[1] == "--debug-profiles":
  for c in range(4):
    print "Mov(%d,%d):" % (c,(c+1)%4)
    for hgp in range(24):
      # On the left, the initial profile
      print "", P4[hgp], " ",
      # On the right, the possible successors
      for hgp2 in tr_prof[hgp][c]:
        print P4[hgp2],
      print
  sys.exit(0)

if len(sys.argv) > 2:
  # Get the weights (at least two) from the command line
  WEIGHTS = [ ]
  for arg in sys.argv[1:]:
    weight = string.atoi(arg)
    # To avoid overflows, make sure they're not too large:
    # clamp the weights in [-1000000,1000000]
    weight = min(weight,1000000)
    weight = max(weight,-1000000)
    WEIGHTS.append(weight)
else:
  # Empty command line?
  # These are exactly the weights needed to prove
  # Proposition 5.3 in the article
  print "%s: using the default weights" % sys.argv[0]
  WEIGHTS = [ 1, 1, 0, -2, -4 ]
DISKS = len(WEIGHTS)-1

# DISKS is the number of primary disks, and N is the number of
# primary configurations. The vertex potentials are stored in
# the bidimensional array P[i][j], where i is the primary
# configuration and j the secondary profile.
# The final weight, i.e. WEIGHTS[-1], is the weight of the
# secondary disks.

N = pow(4,DISKS)
P = [ ]
for vertex in range(N):
  line = [ 0 ] * 24
  P.append(line)

print "Disks: %d" % DISKS
print "Vertices: %d*24 = %d" % (N,N*24)
print "Weights:", WEIGHTS

# An empty column
EMPTY = 999

# Main loop, update the vertex potentials P[i][j] until
# they satisfy all the required inequalities

augmented = 1
while augmented:
  augmented = 0
  for vertex in range(N):
    pos = [ ]
    smallest = [ EMPTY ] * 4
    x = vertex
    for n in range(DISKS):
      x,r = divmod(x,4)
      pos.append(r)
      if n < smallest[r]:
        smallest[r] = n
    # Find all the pairs of adjacent free pegs
    free_pairs = [ ]
    for r in range(4):
      r2 = (r+1) % 4
      if (smallest[r]==EMPTY) and (smallest[r2]==EMPTY):
        # Pegs r and r2 are both free of primary disks, so
        # a secondary disk move would be possible here
        free_pairs.append(r)
    # Horizontal moves (transform the secondary profile)
    line = P[vertex]
    weight = WEIGHTS[-1]
    for hgp in range(24):
      for c in free_pairs:
        for hgp2 in tr_prof[hgp][c]:
          x = line[hgp] - weight
          if x > line[hgp2]:
            line[hgp2] = x
            augmented = augmented+1
    # Vertical moves (move the primary disks)
    for r in range(4):
      r2 = (r+1) % 4
      n = smallest[r]
      if n < smallest[r2]:
        # Disk n can be moved
        weight = WEIGHTS[n]
        offset = (r2-r) * pow(4,n)
        newl = P[vertex+offset]
        for hgp in range(24):
          x = line[hgp] - weight
          if x > newl[hgp]:
            newl[hgp] = x
            augmented = augmented+1
    # Loop to next vertex
  # All vertices explored, check the new values of P
  minp = maxp = P[0][0]
  for line in P:
    minp = min(minp,min(line))
    maxp = max(maxp,max(line))
  print "min(P)=%d, max(P)=%d, aug=%d" % (minp,maxp,augmented)
  if maxp>1000000000:
    # Actually, we have a very comfortable margin here,
    # even on 32-bit machines
    print "Risk of integer overflow, abort"
    sys.exit(8)
  if minp>0:
    # Obstruction found
    print "Strictly negative cycle encountered, abort"
    sys.exit(1)
  # And loop again (while augmented>0)
  
# No more augmentations, we have finished
print "All cycles positive, good"

# Just to be sure
print "Checking symmetries:",

# How do the profiles transform by rotation?
rot_prof = [ ]
for hgp in range(24):
  # Add 1 mod 4 to all the column numbers
  hgp2 = ""
  for c in P4[hgp]:
    c = string.atoi(c)
    c2 = (c+1) % 4
    hgp2 = hgp2 + str(c2)
  hgp2 = P4.index(hgp2)
  rot_prof.append(hgp2)

for vertex in range(N):
  x = vertex
  newv = 0
  for n in range(DISKS):
    x,r = divmod(x,4)
    # Shift all disks to the right
    r2 = (r+1) % 4
    newv = newv + r2 * pow(4,n)
  for hgp in range(24):
    # Shift the profile too
    hgp2 = rot_prof[hgp]
    oldval = P[vertex][hgp]
    newval = P[newv][hgp2]
    if (oldval != newval):
      # This should never happen
      print "symmetry broken, check the program"
      sys.exit(16)
print "P is Z/4Z-symmetric"

# Normal termination (success)

print "We have proved a theorem:"
for i in range(DISKS+1):
  w = WEIGHTS[i]
  plus = ""
  if (w >= 0):
    plus = "+"
  andmore = ""
  if (i == DISKS):
    andmore = "+"
  print "%s%d.x[%d%s]" % (plus,w, i,andmore),
print ">= 0."

utime,stime,cutime,cstime,elapsed = os.times()
print "Time used: %.3f sec" % (utime+stime+cutime+cstime)
