#!/usr/bin/env python
#
# $Id: zmann.py,v 1.7 2020/11/16 16:11:01 bousch Exp $
#
# Computation of Zimmermann's constant Z, the unique real number
# such that Z^3 = Z^2+2

# It can be readily computed using Cardan's formula.
# The following function returns Z as a float,
# with approximately 16 correct digits.

def zmann():
  from math import sqrt, pow
  s = pow(28+sqrt(27*29), 1/3.)
  z = (1+s+1/s)/3
  return z

# For more precision, you can use the following routines.

# Returns the integer part of N*Z, where N>0 is an integer
# or (preferably) a long integer:

def mult_zmann(N):
  # The initial interval [x,y] is [N,2N], and in this interval
  # we have to solve the equation z*z*(z-N) == 2*N*N*N
  x = N
  y = 2*N
  target = 2*N*N*N
  # Dichotomy loop: reduce the interval [x,y]
  while y-x > 1:
    z = (x+y) / 2
    if z*z*(z-N) > target:
      # z is too large, new interval is [x,z]
      y = z
    else:
      # z is too small, new interval is [z,y]
      x = z
    # and loop again
  # After exiting the loop, we have y == x+1
  return x

# Returns the decimal value of Z (as a string), with a given number
# of digits after the decimal point, rounded below

def dec_string(digits):
  N = pow(10L,digits)
  S = mult_zmann(N)
  # Convert the long int to a string (without the trailing 'L')
  S = repr(S)[:-1]
  # And insert the decimal point
  return S[0] + "." + S[1:]

# Writes a python comment like the one just below.

def write_comment(lines):
  lines = max(1,lines)
  print "# The first %d digits of Z are" % (1+50*lines)
  print "#"
  S = dec_string(50*lines)
  for i in range(lines):
    line = "#   "
    if i==0:
      # Integer part of Z
      line = line + S[:2]
      S = S[2:]
    else:
      # Just for alignment
      line = line + "  "
    for j in range(10):
      line = line + S[:5] + " "
      S = S[5:]
    # Remove the trailing space
    line = line[:-1]
    print line
  return

# The first 201 digits of Z are
#
#   1.69562 07695 59862 05741 63671 00117 53534 26181 79388 20850
#     77302 21870 72844 52445 34540 80072 21399 66879 55756 81990
#     06348 36413 49654 66572 12721 10825 18839 00772 59239 42438
#     28691 47845 23487 83163 62889 76173 55955 09659 61141 30961

def main():
  from sys import argv
  from string import atoi
  # Write some lines (the default is 4) giving the first figures of Z,
  # as a python comment.
  # With an empty command line, you should get the comment above.
  try:
    lines = atoi(argv[1])
  except:
    # Empty or malformed command line
    lines = 4
  write_comment(lines)

if __name__ == "__main__":
  main()
